#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon May 18 18:44:09 2020

@author: mario
"""
import networkx as nx
import gurobipy as gp
from gurobipy import GRB
import datetime
import sys
import matplotlib.pyplot as plt
#import lp_to_nx_graph
from itertools import combinations
from enum import Enum

class ConnectivityConstraint(Enum):
    SEPARATORS = 1
    MILLER_TUCKER_ZEMLIN = 3
    MARTIN = 4
    
class KHopDominatingSet():
    
    def __init__(self, G, k, name = "kDS"):
        self.G = G
        self.m = gp.Model(name)
        self.nodes = self.m.addVars(G.nodes, vtype = GRB.BINARY, name = "nodes")
        self.m.setObjective(gp.quicksum(self.nodes), GRB.MINIMIZE)

        # For each node in the graph the node itself or at least one of it's k-neighbors has to be part of the dominating set
        for v in G.nodes: 
            k_neighborhood = set(G.neighbors(v))
            for i in range(k-1):
                k_neighborhood.update([w for n in k_neighborhood for w in G.neighbors(n)])
            self.m.addConstr((gp.quicksum(self.nodes[n] for n in k_neighborhood) +self.nodes[v]) >= 1 )

    def solve(self):
        # TODO: refactor, maybe use self.nodes directly
        ds = {i for i,x_i in enumerate(self.m.getVars()) if x_i.x == 1}
        return ds
    
    def solve_and_draw(self):
        starttime = datetime.datetime.now()
        ds = self.solve()
        endtime = datetime.datetime.now()
        duration = endtime- starttime
        duration_sec = duration.total_seconds()
        color_map = ['red' if i in ds else 'green' for i in self.G.nodes]
        print(f"duration in seconds: {duration_sec}")
        nx.draw_kamada_kawai(self.G, node_color = color_map, with_labels = True)
        plt.show()


    def make_closure(G, k):
        G_prime = nx.Graph(G)
        for v in G.nodes:
            neighbors = set(G.neighbors(v))
            for i in range(k-1):
                neighbors.update([w for n in neighbors for w in G.neighbors(n)])
            
            G_prime.add_edges_from((v,n) for n in neighbors)
        
        return G_prime

class ConnectedKHopDominatingSet(KHopDominatingSet):
    
    def __init__(self, G, k, name = "CkDS", exclude = {}, constraints = ConnectivityConstraint.SEPARATORS):
        self.G = G
        super().__init__(G, k, name)
        
        self.constraints = constraints
        
        if ConnectivityConstraint.SEPARATORS == constraints:
            if exclude:
                self.m.addConstrs(self.nodes[v] <= gp.quicksum(self.nodes[w] for w in G.neighbors(v)) for v in G.nodes if v not in exclude)
            else:
                self.m.addConstrs(self.nodes[v] <= gp.quicksum(self.nodes[w] for w in G.neighbors(v)) for v in G.nodes)
      
        if ConnectivityConstraint.MILLER_TUCKER_ZEMLIN == constraints:
            # given an undirected graph generates a graph which is bidirectional
            self.G_prime = self.G.to_directed()

            n = len(self.G.nodes)
            self.G_prime.add_nodes_from([n+1, n+2])
            self.G_prime.add_edge(n+1,n+2)
            initial_nodes = self.G.nodes
            
            self.G_prime.add_edges_from((n+1, i) for i in initial_nodes)
            self.G_prime.add_edges_from((n+2, i) for i in initial_nodes)
            self.edges = self.m.addVars(self.G_prime.edges, vtype = GRB.BINARY, name="edges")
            
            self.auxiliary = self.m.addVars(self.G_prime.nodes, vtype= GRB.INTEGER, name="auxiliary")
            
            self.m.addConstr(gp.quicksum(self.edges[(n+2,i)] for i in self.G.nodes) == 1)
            self.m.addConstrs(gp.quicksum(self.edges[(i,j)] for i in self.G_prime.nodes if i != j and (i,j) in self.G_prime.edges) == 1 for j in self.G.nodes)
            
            self.m.addConstrs(self.edges[(n+1,i)] + self.edges[(i,j)] <= 1 for (i,j) in self.G.to_directed().edges)

            self.m.addConstrs((n+1)*self.edges[(i,j)] + self.auxiliary[i] - self.auxiliary[j]+ (n-1)*self.edges[(j,i)] <= n for (i,j) in self.G.to_directed().edges)
            
            A_without_E_and_E_prime = [e for e in self.G_prime.edges if e not in self.G.to_directed().edges]
            self.m.addConstrs((n+1)*self.edges[(i,j)]+self.auxiliary[i]-self.auxiliary[j] <= n for (i,j) in A_without_E_and_E_prime)
            
            self.m.addConstr(self.edges[(n+1,n+2)] == 1)
            
            self.m.addConstr(self.auxiliary[n+1] == 0)
            self.m.addConstrs(self.auxiliary[i] >= 1 for i in self.G_prime.nodes if i != n+1)
            self.m.addConstrs(self.auxiliary[i] <= n+1 for i in self.G_prime.nodes if i != n+1)
            
            self.m.addConstrs(self.nodes[i] == 1-self.edges[(n+1,i)] for i in self.G.nodes)
            
        if ConnectivityConstraint.MARTIN == constraints:
            # large integer
            M = 99999
            
            E = [self.G.edges]
            
            self.edges = self.m.addVars(((i,j) for i in self.G.nodes for j in self.G.nodes), vtype = GRB.BINARY, name = "edges")
            
            self.z = self.m.addVars( (((i,j) ,k) for i in self.G.nodes for j in self.G.nodes for k in self.G.nodes), vtype = GRB.BINARY, name ="z")
            
            self.m.addConstr(gp.quicksum(self.edges) == gp.quicksum(self.nodes) - 1)
            
            self.m.addConstrs(self.edges[(i,j)] <= self.nodes[i] for (i,j) in self.G.edges)
            self.m.addConstrs(self.edges[(i,j)] <= self.nodes[j] for (i,j) in self.G.edges)
            
            self.m.addConstrs(self.z[((i,j),k)] <= self.edges[(i,j)] for (i,j) in self.G.edges for k in self.G.nodes)
            self.m.addConstrs(self.z[((j,i),k)] <= self.nodes[k] for (i,j) in self.G.edges for k in self.G.nodes)
            
            self.m.addConstrs(self.edges[(i,j)] - M*(3-self.nodes[i]-self.nodes[j]-self.nodes[k]) <= self.z[((i,j),k)] + self.z[((j,i),k)] for i in self.G.nodes for j in self.G.nodes for k in self.G.nodes)
            self.m.addConstrs(self.z[((i,j),k)] + self.z[((j,i),k)] <= self.edges[(i,j)] +M*(3-self.nodes[i]-self.nodes[j]-self.nodes[k]) for i in self.G.nodes for j in self.G.nodes for k in self.G.nodes)
                                    
            self.m.addConstrs(1-M*(2-self.nodes[i]-self.nodes[j]) <= gp.quicksum(self.z[((i,k),j)] for k in self.G.nodes if k != i and k != j if (i,k) in self.G.to_directed().edges)  +self.edges[(i,j)] for i in self.G.nodes for j in self.G.nodes)
            
            self.m.addConstrs(self.edges[(i,j)] == 0 for i in self.G.nodes for j in self.G.nodes if (i,j) not in E)
            self.m.addConstrs(self.z[((i,j),k)] == 0 for i in self.G.nodes for j in self.G.nodes for k in self.G.nodes if (i,j) not in E)
            
            
    def min_ij_separator(G, i, j, C_i):
        N_ci = {v for c in C_i for v in G.neighbors(c)}
        G_prime = nx.Graph(G)
        C_i_prime = C_i.copy()
        C_i_prime.update(N_ci)
        G_prime.remove_edges_from(G.subgraph(C_i_prime).edges)
        # dijkstra
        R_j = nx.algorithms.dag.descendants(G_prime, j)
        return R_j.intersection(N_ci)
    
    def solve(self):

        if ConnectivityConstraint.SEPARATORS == self.constraints:
            self.m._vars = self.nodes
            self.m._G = self.G
            self.m.Params.lazyConstraints = 1
            self.m.optimize(ConnectedKHopDominatingSet.elim_unconnectivity)    

        elif ConnectivityConstraint.INDEGREE == self.constraints:
            self.m.optimize()
        elif ConnectivityConstraint.MILLER_TUCKER_ZEMLIN == self.constraints:
            self.m.optimize()
        elif ConnectivityConstraint.MARTIN == self.constraints:
            self.m.optimize()
        
        ds = {i for i,x_i in enumerate(self.m.getVars()) if x_i.x > 0.5}
        return ds
    
    # Probably uncorrect version
    # def elim_unconnectivity(model, where):
    #     if where == GRB.Callback.MIPSOL:
    #         vals = model.cbGetSolution(model._vars)
    #         ds = {i for i in model._vars.keys() if vals[i] > 0.5}
            
    #         G_prime_prime = model._G.subgraph(ds)
    #         if(not nx.is_connected(G_prime_prime)):
    #             C = [c for c in nx.algorithms.components.connected_components(G_prime_prime)]
    #             for i in range(len(C)-1):
    #                 C_i = C[i]
    #                 for j in range(i+1, len(C)):
    #                     C_j = C[j]
    #                     h = next(iter(C_i))
    #                     l = next(iter(C_j))
    #                     min_ij_sep = ConnectedKHopDominatingSet.min_ij_separator(model._G, h, l, C_i)
    #                     model.cbLazy(gp.quicksum(model._vars[s] for s in min_ij_sep) >= model._vars[h] + model._vars[l] - 1)

    # V1  
    # def elim_unconnectivity(model, where):
    #     if where == GRB.Callback.MIPSOL:
    #         vals = model.cbGetSolution(model._vars)
    #         ds = {i for i in model._vars.keys() if vals[i] > 0.5}
            
    #         G_prime_prime = model._G.subgraph(ds)
    #         if(not nx.is_connected(G_prime_prime)):
    #             C = [c for c in nx.algorithms.components.connected_components(G_prime_prime)]
    #             for i in range(len(C)):
    #                 C_i = C[i]
    #                 for j in range(len(C)):
    #                     C_j = C[j]
    #                     h = next(iter(C_i))
    #                     l = next(iter(C_j))
    #                     min_ij_sep = ConnectedKHopDominatingSet.min_ij_separator(model._G, h, l, C_i)
    #                     if min_ij_sep:
    #                         model.cbLazy(gp.quicksum(model._vars[s] for s in min_ij_sep) >= model._vars[h] + model._vars[l] - 1)
                            
    # V2
    def elim_unconnectivity(model, where):
        if where == GRB.Callback.MIPSOL:
            vals = model.cbGetSolution(model._vars)
            ds = {i for i in model._vars.keys() if vals[i] > 0.5}
            
            G_prime_prime = model._G.subgraph(ds)
            if(not nx.is_connected(G_prime_prime)):
                C = [c for c in nx.algorithms.components.connected_components(G_prime_prime)]
                for i in range(len(C)):
                    C_i = C[i]
                    for j in range(len(C)):
                        C_j = C[j]
                        h = next(iter(C_i))
                        l = next(iter(C_j))
                        min_ij_sep = ConnectedKHopDominatingSet.min_ij_separator(model._G, h, l, C_i)
                        if min_ij_sep:
                            for w in C_i:
                                model.cbLazy(gp.quicksum(model._vars[s] for s in min_ij_sep) >= model._vars[w] + model._vars[l] - 1)
        
class RootedConnectecKHopDominatingSet(ConnectedKHopDominatingSet):
    
    def __init__(self, G, k, root = 0, name = "RCkDS", constraints = ConnectivityConstraint.SEPARATORS):
        super().__init__(G, k, name, exclude = {root}, constraints = constraints)
        self.root = root
        
        self.m.addConstr(self.nodes[root] >= 1)
        
        # self.add_all_combinations_root_separators()
        # print("neighborhood separators:")
        # self.add_all_neighborhood_root_separators()
        # print("single node separators:")
        # self.add_single_root_separators()
        
    def add_single_root_separators(self):
        number_of_found_separators = 0
        for i in self.G.nodes:
            min_ij_sep = ConnectedKHopDominatingSet.min_ij_separator(G, i, self.root, {i})
            if min_ij_sep:
                self.m.addConstr(gp.quicksum(self.nodes[s] for s in min_ij_sep) >= self.nodes[i])
                number_of_found_separators += 1
        print(f"number of preadded separators: {number_of_found_separators}")
        
    # ToDo: refactor!
    def add_all_neighborhood_root_separators(self):
        number_of_found_separators = 0
        for v in self.G.nodes:
            # V1
            to_add = set()
            if v is not self.root and self.root not in self.G.neighbors(v) and v not in self.G.neighbors(self.root) and not set(self.G.neighbors(self.root)).intersection(set(self.G.neighbors(v))):
                for i in range(2,self.G.degree[v]+1):
                    V = {w for w in self.G.neighbors(v)}
                    V.update([v])
                    for i_neighborhood in combinations(V, i):
                        # V3
                        # to_add = set()
                        if v in i_neighborhood:
                            # V2
                            # number_of_found_separators += 1
                            min_ij_sep = ConnectedKHopDominatingSet.min_ij_separator(self.G, v, self.root, set(i_neighborhood))
                            # V1, V3
                            to_add.add(frozenset(min_ij_sep))
                            
                            # V2
                            # self.m.addConstr(gp.quicksum(self.nodes[s] for s in min_ij_sep) >= self.nodes[v])
                        
                            # V3 
                            # number_of_found_separators += len(i_neighborhood)
                        # V3
                        # self.m.addConstrs(gp.quicksum(self.nodes[s] for s in min_ij_sep_to_add) >= self.nodes[w] for w in i_neighborhood for min_ij_sep_to_add in to_add)
            # V1
            self.m.addConstrs(gp.quicksum(self.nodes[s] for s in min_ij_sep) >= self.nodes[v] for min_ij_sep in to_add)
            # V1
            number_of_found_separators += len(to_add)
            
            print(f"number of preadded separators: {number_of_found_separators}")
        
    
    def add_simple_path_length_constraint(self):
        self.m.addConstrs((self.nodes[v] * len(nx.algorithms.shortest_path(self.G, self.root, v))) <= gp.quicksum(self.nodes) for v in self.G.nodes)
        # self.m.addConstrs((self.nodes[v] * self.nodes[w] * len(nx.algorithms.shortest_path(self.G, v, w))) <= gp.quicksum(self.nodes) for v in self.G.nodes for w in self.G.nodes)

        
    def add_gausian_sum_formula_constraint(self):
        self.m.addConstr(gp.quicksum(self.nodes[v] * len(nx.algorithms.shortest_path(self.G, self.root, v)) for v in self.G.nodes) <= (gp.quicksum(self.nodes)+1)*gp.quicksum(self.nodes)/2)
        # self.m.addConstr(gp.quicksum(self.nodes[v] * self.nodes[w] * len(nx.algorithms.shortest_path(self.G, v, w)) for v in self.G.nodes for w in self.G.nodes) <= (gp.quicksum(self.nodes)+1)*gp.quicksum(self.nodes)/2)

    def add_neighbor_of_neighbors_constraint(self, exclude):
        self.m.addConstrs(self.nodes[v] <= gp.quicksum(self.nodes[w] * gp.quicksum(self.nodes[h] for h in self.G.neighbors(w) if h is not v) for w in self.G.neighbors(v)) for v in self.G.nodes if v not in exclude)

    def intermediate_node_constraint(self, exclude):
        self.m.addConstrs(2 * self.nodes[v] <= gp.quicksum(self.nodes[w] for w in G.neighbors(v)) for v in G.nodes if v not in exclude)        


if __name__ == '__main__':
    G = lp_to_nx_graph.read(sys.argv[1])
    G = G.to_undirected()
    
    if(len(sys.argv) > 2):
        k = int(sys.argv[2])
    else:
        k = 1
        
    
    # G = lp_to_nx_graph.read("/home/mario/Dokumente/Uni/Bachelorarbeit/git_repo/bachelor-mario-surlemont/python/lp graphs/small-leaf.lp")
    # k = 1
    dsProb = RootedConnectecKHopDominatingSet(G, k, 0, constraints = ConnectivityConstraint.SEPARATORS)
    dsProb.solve_and_draw()
